# written by Anirban Bhattacharya
# do not share without permission




# function to fit Bayesian ridge regression
# Model: y ~ N(Xbeta,sigma^2 I_n)
# prior: beta | sigma^2 ~ N(0, sigma^2 xi^(-1) I_d)
#               sigma^2 ~ IG(a_0/2,b_0/2)
#                  tau  ~ Ca_+(0,1), where tau = 1/sqrt(xi) 


# sampling steps: (i) sample xi using MH random walk on log(xi)
# (ii) sigma^2 | xi, and (iii) beta | sigma^2, xi


BayesRidge = function(y,X,sig_MH,a0=0.5,b0=0.5,BURNIN=1000,MCMC=10000,THIN=5){
  # y is nx1 response, X is nxd design matrix
  # sig_MH > 0 is the constant in the sigmoidal function approximating the indicator
  # a0, b0 are the prior IG hyperparameters (set to 0 for improper prior)
  
  # create various design matrix stuff outside MCMC loop
  n = nrow(X)
  d = ncol(X)
  I_n=diag(n)
  I_d = diag(d)
  
  sX = svd(X)$d     # vector of singular values of X
  G = X%*%t(X)      # Gram matrix XX'
  H = t(X)%*%X      # X'X
  Xdoty = t(X)%*%y  # X'y
  
  # global MCMC stuff
  NITER = BURNIN + MCMC
  EFF = MCMC/THIN
  
  # output files  
  betaout = matrix(0,nrow=EFF,ncol=d)  
  sigmasqout = rep(0,EFF)
  xiout = rep(0,EFF)

  # initialize
  xi = 1
  # M = I_n + xi^(-1)XX' (don't actually create it)
  lgdetM = sum(log(1 + (1/xi)*(sX^2)))   # log(det(M))
  qfy = t(y)%*%solve(I_n+(1/xi)*G,y)     # y'M^(-1)y
  lgpost = lgmarglikxi(xi,lgdetM,qfy,a0,b0) + lgpixi(xi) + log(xi)

  
  # begin MCMC
  for(g in 1:NITER) {
    
    # -- update xi -- #
    lgxi = log(xi)
    lgxiprop = lgxi + sig_MH*rnorm(1)
    xiprop = exp(lgxiprop)
    lgdetMprop = sum(log(1 + (1/xiprop)*(sX^2)))   
    qfyprop = t(y)%*%solve(I_n+(1/xiprop)*G,y)
    lgpostprop = lgmarglikxi(xiprop,lgdetMprop,qfyprop,a0,b0) + lgpixi(xiprop) + log(xiprop)
    
    MHrat = min(1,exp(lgpostprop-lgpost))
    uu = runif(1)
    
    if (uu < MHrat){
      xi = xiprop
      # lgdetM = lgdetMprop    #(not used later)
      qfy = qfyprop
      lgpost = lgpostprop
    }
    
    
    # -- update sigmasq -- #
    sigmasq = 1/rgamma(1,shape=0.5*(n+a0),rate=0.5*(qfy+b0))
    
    # -- update beta -- #
    beta = mvnrue((1/sigmasq)*(H+xi*I_d),(1/sigmasq)*Xdoty)
    
    
    # -- save stuff -- #
    if(g > BURNIN && g%%THIN==0) {
      betaout[(g-BURNIN)/THIN,] = beta
      sigmasqout[(g-BURNIN)/THIN] = sigmasq
      xiout[(g-BURNIN)/THIN] = xi
    }
    
  }
  postmean = apply(betaout,2,mean)
  return(list('Betaout' = betaout,'postmean_beta' = postmean,'Xiout' = xiout,'Sigout' = sigmasqout)) 
  
  }
  

